void logicle_scale(double t, double w, double m, double a, double* x, int n);
void logicle_inverse(double t, double w, double m, double a, double* x, int n);
void hyperlog_scale(double t, double w, double m, double a, double* x, int n);
void hyperlog_inverse(double t, double w, double m, double a, double* x, int n);